//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* list.cpp                                                               *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.messen-und-deuten.de/frida                                  *//
//**************************************************************************//

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <algorithm>
using namespace std;

#include "mystd.h"
#include "list.h"


//**************************************************************************//
//*  Constructors                                                          *//
//**************************************************************************//

//! Initialize undefined list.

CList::CList() :
    instruction( "-" ), limgiven( false )
{
    parse( instruction );
}

//! Initialize by string.

CList::CList( string inp ) :
    instruction( inp ), limgiven( false )
{
    parse( instruction );
}

//! Set limits and initialize by string.

CList::CList( string inp, uint _inf, uint _sup ) :
    instruction( inp ), inf( _inf), sup( _sup ), limgiven( true )
{
    parse( instruction );
}

//! Initialize list containing one value.

CList::CList( uint val ) :
    instruction(""), limgiven( false ), limneeded( false )
{
    V.push_back(val);
}

//! Initialize list containing one interval. Do not set limits.

CList::CList( uint from, uint to ) :
    limgiven( false )
{
    if( to == (uint)(from-1) ){
        instruction = "-";
        parse( instruction );
    } else {
        if ( to<from || to==(uint)-1 )
            throw "abusive CList initialization (from " + 
                strg(from) + " to " + strg(to) + ")";
        instruction = "";
        limneeded = 0;
        for (uint val = from; val <= to; ++val)
            V.push_back(val);
    }
}


//**************************************************************************//
//*  Interactive initialization (broken)                                   *//
//**************************************************************************//


/* 
void CList::ask_and_set( const string& quest )
{
}
input help:
            cout << "required input type: index list (without blanks)\n"
                "examples:\n"
                "    0,2,7         comma-separated list of values\n"
                "    0-7           range\n"
                "    0-7.2         each 2nd: 0,2,4,6\n"
                "    -             empty list\n"
                "    *             full range, according to context\n"
                "    8-*           from 8 to end of range\n";
*/

//**************************************************************************//
//*  Manipulate list                                                       *//
//**************************************************************************//

//! Set bounds and evaluate instruction.

void CList::evaluate( uint _inf, uint _sup )
{
    inf = _inf;
    sup = _sup;
    limgiven = true;
    parse(instruction);
}

//! Re-arrange V (normally, the user-supplied order is retained).

void CList::sort(void)
{
    ::sort( V.begin(), V.end() );
    instruction = "";
}

void CList::sortReverse(void)
{
    sort();
    reverse( V.begin(), V.end() );
    instruction = "";
}

//! Append other sequence or item, unless duplicate.

void CList::append( const CList L )
{
    instruction = "";
    for ( uint i=0; i<L.V.size(); ++i )
        if ( !contains( L.V[i] ) )
            V.push_back( L.V[i] );
}

void CList::append( const uint val )
{
    instruction = "";
    if ( !contains(val) )
        V.push_back(val);
}


//**************************************************************************//
//*  Inspect                                                               *//
//**************************************************************************//

bool CList::contains(uint val) const
{
    for (uint i=0; i<V.size(); i++)
        if (val==V[i]) return 1;
    return 0;
}

bool CList::all_between( const int imi, const int ima )
{
    for ( uint i=0; i<size(); i++ )
        if ( V[i]<imi || V[i]>ima )
            return false;
    return true;
}

string CList::str(void) const
{ 
    if (instruction != "") return instruction;
    return sel2text();
}

string CList::sel2text(void) const
{
    if(!V.size())
        return "-";
    string s;
    vector<uint> VV=V;
    vector<uint>::iterator iV=VV.begin();
    uint vi, vd;
    while(iV<VV.end()) {
        s += strg(vi=*(iV++));
        if(iV>=VV.end()) break;
        vd = *(iV++) - vi;
        if(iV>=VV.end()) {
            s += "," + strg(vi+vd);
            break;
        }
        if (*iV!=vi+2*vd) {
            s += ",";
            iV--;
            continue;
        }
        while(iV<VV.end() && *(iV-1)+vd==*iV) iV++;
        s += "-" + strg(*(iV-1));
        if (vd!=1) s += "." + strg(vd);
        if(iV>=VV.end()) break;
        s += ",";
    }
    return s;
}


//**************************************************************************//
//*  parse string                                                          *//
//**************************************************************************//

void CList::parse( string inp )
{
    instruction = inp;
    V.clear();
    if( inp.substr(0,1)=="/" ){
        // handle this case by recurring to complementary list
        limneeded = 1;
        if (!limgiven)
            return; // cannot evaluate until limits are given
        CList VCompl( inp.substr(1), inf, sup );
        for( uint i=inf; i<=sup; ++i )
            if( !VCompl.contains(i) )
                V.push_back( i );
    } else {
        limneeded = 0;
        if(inp==string("-")) return; // valid empty list
        string::size_type ji=0, jf;
        while (ji<inp.size()) {
            if ((jf=inp.find(',', ji)) == string::npos) jf = inp.size();
            parseSkipinterval(inp.substr(ji,jf-ji));
            ji = jf+1;
        }
    }
}

void CList::parseSkipinterval( string inp )
{
    int skip;
    string::size_type j;
    if ((j=inp.find('.')) != string::npos) {
        skip = parseSkipfactor(inp.substr(j+1));
    } else {
        skip = 1;
        j = inp.size();
    }
    parseInterval(inp.substr(0,j), skip);
}

int CList::parseSkipfactor( string inp )
{
    int val;
    char dummy;
    if (sscanf(inp.c_str(), "%d%c", &val, &dummy)!=1 || val<=0)
        throw "invalid skip factor [" + inp + "]";
    return val;
}

void CList::parseInterval( string inp, int skip )
{
    string::size_type j;
    uint vi, vf;
    bool xv=0; // vf exists ?
    if (inp==string("*")) {
        limneeded = 1;
        if (!limgiven)
            return; // cannot evaluate until limits are given
        vi = inf;
        vf = sup;
        xv = 1;
    } else if ((j=inp.find('-')) != string::npos) {
        vf = parseListvalue(inp.substr(j+1),+1);
        vi = parseListvalue(inp.substr(0,j),-1);
        xv = 1;
    } else {
        vi = parseListvalue(inp,0);
        vf = vi;
    }
    if( limneeded && !limgiven )
            return; // cannot evaluate until limits are given
    if(vi>vf)
        throw "invalid interval [" + inp + "]";
    for (uint vv=vi; vv<=vf; vv += skip)
        V.push_back(vv);
}	

int CList::parseListvalue( string inp, int which )
{
    if (inp==string("*")) {
        limneeded = 1;
        if (!limgiven)
            return -1; // cannot evaluate until limits are given
        if      (which==-1) 
            return inf;
        else if (which==+1) 
            return sup;
        else
            throw string( "BUG: list value * out of context" );
    }
    int val;
    char dummy;
    if (sscanf(inp.c_str(), "%d%c", &val, &dummy)!=1 || val<0)
        throw "invalid list value [" + inp + "]";
    if ( limgiven && (val<inf || val>sup || sup==(uint)(-1)) )
        throw "list value [" + inp + "] out of bounds ("
            + strg(inf) + "," + strg((int)sup) + ")";
    return val;
}
